/* Copyright (c) 1997, MpegTV LLC */

#ifndef lint
static char rcsid[] = "$Header: /net/hopkins/var/home/tristan/mpeg/dec/RCS/ci_file.c,v 1.2 1997/08/21 11:51:26 tristan Exp $";
static char copyright[] = "Copyright (c) 1997, MpegTV LLC";
#endif

/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any license fee or
 * royalty on an "as is" basis.  Copyright-holder disclaims
 * any and all warranties, whether express, implied, or statuary, including any
 * implied warranties or merchantability or of fitness for a particular
 * purpose.  In no event shall the copyright-holder be liable for any
 * incidental, punitive, or consequential damages of any kind whatsoever
 * arising from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and user's
 * customers, employees, agents, transferees, successors, and assigns.
 *
 */
 
/***************************************************************************

This is an example of a VCI (Video Capture Interface) plugin.

A VCI plugin is a dynamic library that an application can
dynamically link to mtvp to extend it.

A VCI plugin can be linked to mtvp using the COM_VCI_LINK
command.

A VCI plugin should define the following routines:

int _init(void): called right after the plugin is linked
int _fini(void): called right before the plugin is unlinked
int VCI_notify(const YUV_IMAGE *image): called right before a frame is displayed
int VCI_cntl(const void *data, u_int size): called when a COM_VCI_CNTL message is sent

A VCI plugin can send COM_A2G_VCI_FEEDBACK messages to the
application by calling vci_feedback:

int vci_feedback(const void *data, u_int size)

This particular file is a working example that is used by the
sample program test_vci to capture frames while playing
video in real time. This very same example of VCI could also
be used to capture frames not in real time, by the way...

Each time a frame is to be displayed by mtvp, right before
it is converted to the appropriate format for the display
(a process sometimes called dithering), if a VCI plugin
is linked, mtvp calls VCI_notify:

int VCI_notify(const YUV_IMAGE *image)

In this example, VCI_notify captures the frame by
copying it in a segment of shared memory.  The captured
frame is then processed asynchronously by the controling
application.  A semaphore is also shared between this
VCI and the application to synchronize access to the
shared memory segment.

In this example, Both the shared memory segment
and the semaphore are allocated by the VCI plugin and the
corresponding Ids are passed to the application using
the vci_feedback routine.

Note that in this example the application is responsible
for deleting the shared memory segment and the semaphore.

In this example, the application is notified that a frame is available
in the shared memory segment by sending it a zero-size message
send with vci_feedback.

***************************************************************************/


#include "mtvp/misc.h"
#include "mtvp/vci_api.h"

#include "sample_vci.h"


/*--------------------------------------------------------------------------+
| Statics																	|
+--------------------------------------------------------------------------*/


static SHARED_IDS Shared_ids;

static Bool Init_notify;

static Bool	Capture_frame;


/*--------------------------------------------------------------------------+
| _init																		|
+--------------------------------------------------------------------------*/
int
_init()
{
	/*
	 * If the library exports a routine named  _init(),  then  that
     * code  is  executed  before  dlopen()  returns.
     */

#ifdef DEBUG
	printf("_init(): loading VCI\n");
#endif
	

	Shared_ids.shmid[0] = Shared_ids.shmid[1] = Shared_ids.shmid[2] = -1;		

	return 0;
}


/*--------------------------------------------------------------------------+
| _fini																		|
+--------------------------------------------------------------------------*/
int
_fini()
{
	/*
	 * If the dynamic library exports a routine named _fini(),
	 * then that routine is called just before the library is unloaded.
	 */

#ifdef DEBUG
	printf("_fini(): unloading VCI\n");
#endif
		
	return 0;
}




/*--------------------------------------------------------------------------+
| VCI_notify																|
+--------------------------------------------------------------------------*/
int
VCI_notify(image)
	const YUV_IMAGE *image;
{	
	/*
	 * Called for each frame to be displayed.
	 */
	 
	u_int	n;
		
#ifdef DEBUG
	printf("VCI_notify: frame %u, %s\n", image->frame_infos.display_index, Capture_frame ? "captured" : "skipped");
#endif

	if (!Capture_frame) {
		return 0;
	}
		
	{
		const YUV_IMAGE_AND_INDEX *p = (const YUV_IMAGE_AND_INDEX *)image->plane[(int)Y].array;
		
		n = (p - 1)->index;
	}
	
#ifdef DEBUG
	printf("VCI_notify: frame buffer index %u\n", n);
#endif

	/* send shared info to application */
	if (!Init_notify) {
		/*
		 * This block of code is executed only the first time
		 * VCI_notify() is called.
		 */		 

		/* write down frame size */
		Shared_ids.width = image->plane[(int)Y].width;
		Shared_ids.height = image->plane[(int)Y].height;

#ifdef DEBUG
		printf("VCI_notify: width = %u, height = %u\n", Shared_ids.width, Shared_ids.height);
#endif

		/* send the shared Ids to the application */
		if (vci_feedback((const char *)&Shared_ids, sizeof Shared_ids) != 0) {
			perror("VCI_notify: vci_feedback");
			abort_now();
		}

		Init_notify = True;
	}


	/* copy YUV_IMAGE structure */
	*(YUV_IMAGE *)((const char *)image->plane[(int)Y].array - sizeof(YUV_IMAGE_AND_INDEX)) = *image;

	Capture_frame = False;

	/*
	 * Notify application that a frame is read to be read from the
	 * shared memory segment by sending a zero-size message, for example.
	 */
	vci_feedback(&n, sizeof n);


	/* VCI_notify should return 0 */
	return 0;
}

/*--------------------------------------------------------------------------+
| VCI_cntl																	|
+--------------------------------------------------------------------------*/
int
VCI_cntl(data, size)
	const void *data;
	u_int		size;
{
#ifdef DEBUG
	printf("VCI_cntl(data, %u)\n", size);
#endif

	/*
	 * The application sends an empty COM_VCI_CNTL message to tell
	 * the VCI plugin to grab the next frame.
	 */
	Capture_frame = True;
	
	return 0;
}


/*--------------------------------------------------------------------------+
| VCI_alloc																	|
+--------------------------------------------------------------------------*/
void *
VCI_alloc(size)
	size_t	size;
{
	void *shared_mem;
	u_int		n;
	
	
#ifdef DEBUG
	printf("VCI_alloc(%u)\n", size);
#endif

	/* VCI_alloc is only called 3 times */
	if (Shared_ids.shmid[0] == -1) {
		n = 0;
	} else if (Shared_ids.shmid[1] == -1) {
		n = 1;
	} else if (Shared_ids.shmid[2] == -1) {
		n = 2;
	} else {
		return NULL;
	} 

	/*
	 * Allocate a shared memory segment descriptor large enough to
	 * fit the YUV_IMAGE_AND_INDEX structure followed by the frame planes.
	 */
	size += sizeof(YUV_IMAGE_AND_INDEX);
#ifdef DEBUG
	printf("size = %u\n", size);
#endif

	/* create the shared memory segment and get the shared memory segment id */
	Shared_ids.shmid[n] = shmget(IPC_PRIVATE, size, IPC_CREAT|IPC_EXCL|ACCESS_PERMS);
	if (Shared_ids.shmid[n] == -1) {
		perror("VCI_notify: shmget");
		return NULL;
	}
#ifdef DEBUG
	printf("Shared_ids.shmid[%u] = %d\n", n, Shared_ids.shmid[n]);
#endif

	/* attach shared memory segment */
	if ((shared_mem = shmat(Shared_ids.shmid[n], NULL, ACCESS_PERMS)) == (char *)-1) {
		perror("VCI_notify: shmat");
		return NULL;
	}

	/* copy the index in the YUV_IMAGE_AND_INDEX at the beginning of the segment */
	{
		YUV_IMAGE_AND_INDEX *p = (YUV_IMAGE_AND_INDEX *)shared_mem;
		
		p->index = n;
	}
	
	/* return the address where the plane data will be stored */
	return shared_mem + sizeof(YUV_IMAGE_AND_INDEX);
}

/*--------------------------------------------------------------------------+
| VCI_free																	|
+--------------------------------------------------------------------------*/
void
VCI_free(data)
	void	*data;
{
	u_int		n;
	
#ifdef DEBUG
	printf("VCI_free(data)\n");
#endif

	{
		const YUV_IMAGE_AND_INDEX *p = (const YUV_IMAGE_AND_INDEX *)data;
		
		n = (p - 1)->index;
	}

	/* make sure the segment will be destroyed once detached */
#ifdef DEBUG
	printf("removing Shared_ids.shmid[%u] = %u\n", n, Shared_ids.shmid[n]);
#endif
	if (shmctl(Shared_ids.shmid[n], IPC_RMID, NULL) == -1) {
		perror("VCI_free: shmctl");
	} else {
		Shared_ids.shmid[n] = -1;
	}

	/* detach segment */
#ifdef DEBUG
	printf("detaching shared memory segment\n");
#endif
	if (shmdt((char *)data - sizeof(YUV_IMAGE_AND_INDEX)) == -1) {
		perror("VCI_free: shmdt");
	}

}

/* eof */
